/* 
 * Provides the exp_table type, which stores the expansions of a word.
 * Use it like this:
 * 
 *     int i;
 *     struct exp_table t;
 *     exp_table_init (&t, "paint");
 *     add_expansion_copy (&t, "painted", mask0);
 *     add_expansion_copy (&t, "painting", mask1);
 *     add_expansion_copy (&t, "painter", mask2);
 *     for (i = 0; i < num_expansions (&t); i++)
 *         printf("expansion: %s\n", get_expansion (&t, i));
 *     exp_table_empty (&t);
 * 
 * where mask0 is a MASKTYPE with the flags used to add 'ed' set,
 * mask1 gives the flags for 'ing', etc.
 * 
 * Note that allocating the struct itself is up to you, but you
 * should initialize it with exp_table_init() before use and call
 * exp_table_empty() before you free it.
 */

/* The structure itself.  Normally, it is better to use the accessors
 * below rather than access the structure directly.
 */
struct exp_table
  {
  char ** exps;
  MASKTYPE * flags;
  int size;
  const ichar_t * orig_word;
  };

/* Initialize a struct exp_table.  After initialization the number of
 * expansions will be zero.  Pass in the original word from which the
 * expansions are generated - this will be stored by reference. 
 * Returns a pointer to the same struct.
 */
extern struct exp_table * exp_table_init (struct exp_table * e,
					  const ichar_t * orig_name);

/* Return the original word. */
extern const ichar_t * get_orig_word (const struct exp_table * e);

/* Return expansion number i (numbered from zero). */
extern const char * get_expansion (const struct exp_table * e, int i);

/* Return the flags used to get expansion number i. */
extern MASKTYPE get_flags (const struct exp_table * e, int i);

/* Return the number of expansions. */
extern int num_expansions (const struct exp_table * e);

/* Add a new expansion to the list, if it is not already in there.
 * Returns true iff the expansion was added.  Specify the result of
 * the expansion and the flags that were used.  Takes a copy of the
 * string passed in.
 */
extern int add_expansion_copy (struct exp_table * e, const char * s,
			       MASKTYPE flags);

/* Empty the table of expansions, freeing any resources allocated.
 * Returns a pointer to the now empty struct.
 */
extern struct exp_table * exp_table_empty (struct exp_table * e);

/* Dump the contents of a table to stderr, for debugging. */
extern void exp_table_dump (const struct exp_table * e);
