#ifndef lint
static char Rcs_Id[] =
    "$Id: fix8bit.c,v 1.4 2003/08/23 10:56:33 ed Exp $";
#endif

/*
 * Copyright 1993, 1999, 2001, Geoff Kuenning, Claremont, CA
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All modifications to the source code must be clearly marked as
 *    such.  Binary redistributions based on modified source code
 *    must be clearly marked as modified versions in the documentation
 *    and/or other materials provided with the distribution.
 * (clause 4 removed with permission from Geoff Kuenning)
 * 5. The name of Geoff Kuenning may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY GEOFF KUENNING AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL GEOFF KUENNING OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * This is a stupid little program that can be used to convert 8-bit
 * characters to and from backslashed escape sequences.  It is usually
 * more efficient to do this to affix files than to uuencode them for
 * transport.  Ispell will read affix files in either format, so it is
 * merely personal preference as to which form to use.
 *
 * Usage:
 *
 *	fix8bit {-7 | -8} < infile > outfile
 *
 * One of -7 and -8 must be specified.  If -7 is given, any character
 * sequence that is not standard printable ASCII will be converted
 * into a backslashed octal sequence.  If -8 is given, any backslashed
 * octal or hex sequence will be converted into the equivalent 8-bit
 * character.
 *
 * This program is not very smart.  In particular, it makes no attempt
 * to understand comments, quoted strings, or similar constructs in
 * which you might not want conversion to take place.  I suggest that
 * you "diff" the input against the output, and if you don't like the
 * result, correct it by hand.
 */

/*
 * $Log: fix8bit.c,v $
 * Revision 1.4  2003/08/23 10:56:33  ed
 * Purging my old email address.
 *
 * Revision 1.3  2001/11/07 18:59:26  ed
 * Rewrite of fix8bit.c prompted by SuSE's ispell-3.2.06-languages.patch.
 * I wanted to make sure the patch wouldn't break anything, so I wrote a
 * test suite, but doing that I found lots of other things that were
 * wrong, so I started trying to fix those...
 *
 * Makefile: fixed dependencies for fix8bit, added 'test' target.  The
 * test suite checks fix8bit's pushback routines, runs test_fix8bit (see
 * below) and checks a couple of additional properties: fix8bit -8 |
 * fix8bit -7 == cat; fix8bit -7 | fix8bit -7 == fix8bit -7.
 *
 * fix8bit.c: rewrote to8bit() to better handle cases when the
 * backslashed sequence turns out to be illegal.  The initial backslash
 * is printed and the remaining characters are pushed back to be read
 * again.  This means that for example \\x41 will print as \A, in the
 * same way that !\x41 produces !A.  It also handles escape sequences
 * cut off by EOF properly (again they are printed unchanged).  This uses
 * a mini pushback library which has a test suite if you give main() the
 * argument --test-pushback.  Also fixed the original problem with hex
 * sequences being miscomputed, which SuSE wrote the patch for.  Added a
 * warning if the input already contains 8-bit chars (that would stop
 * -8 | -7 being identity).
 *
 * test_fix8bit: new file.  This is a Perl script to run fix8bit -7 and
 * fix8bit -8 on every input file in test_data/ and check the results
 * against the expected results also in that directory.
 *
 * test_data/: new directory.  Contains test cases, some written by hand
 * and some randomly generated by rand_gen.  rand_gen tries to make 'well
 * behaved' input that doesn't muck up fix8bit -8 | fix8bit -7 or other
 * commands - but there are three flags you can use to tell it not to.
 * The random test cases have not been checked by hand, so they should be
 * used in addition to human-written ones.
 *
 * Revision 1.2  2001/10/05 14:22:30  ed
 * Imported 3.2.06.epa1 release.  This was previously developed using
 * sporadic RCS for certain files, but I'm not really bothered about
 * rolling back beyond this release.
 *
 * Revision 1.6  2001/07/25 21:51:47  geoff
 * *** empty log message ***
 *
 * Revision 1.5  2001/07/23 20:43:38  geoff
 * *** empty log message ***
 *
 * Revision 1.4  1999/01/07 06:07:52  geoff
 * Update the copyright.
 *
 * Revision 1.3  1994/01/25  07:12:26  geoff
 * Get rid of all old RCS log lines in preparation for the 3.1 release.
 *
 */

#include <stdio.h>
#include <assert.h>
#include <stdlib.h>

int		main ();	/* Convert to/from 8-bit sequences */
static void	usage ();	/* Issue a usage message */
static void	to7bit ();	/* Convert from 8-bit sequences */
static void	to8bit ();	/* Convert to 8-bit sequences */
static int      get_char ();    /* Get char with pushback */
static void     push_back ();   /* Push back character */
static void     test_pushback ();

extern void	exit ();	/* Terminate program */

int main (argc, argv)		/* Convert to/from 8-bit sequences */
    int			argc;	/* Argument count */
    char *		argv[];	/* Argument vector */
    {
    
    if (argc != 2)
	usage ();
    if (strcmp (argv[1], "-7") == 0)
	to7bit ();
    else if (strcmp (argv[1], "-8") == 0)
	to8bit ();
    else if (strcmp (argv[1], "--test-pushback") == 0)
	{
	/* Magic undocumented test option :-) */
	srand (0);
	test_pushback ();
	}
    else
	usage ();
    return 0;
    }

static void usage ()		/* Issue a usage message */
    {

    (void) fprintf (stderr, "Usage:  fix8bit {-7 | -8} < infile > outfile\n");
    exit (1);
    }

static void to7bit ()		/* Convert from 8-bit sequences */
    {
    int			ch;	/* Next character read from input */

    while ((ch = getchar ()) != EOF)
	{
	ch &= 0xFF;
	if (ch >= 0200)
	    (void) printf ("\\%3.3o", (unsigned) ch);
	else
	    (void) putchar (ch);
	}
    }

/* Simple character input with pushback.  ungetc() handles only a
   single character, unfortunately. */
#define MAX_PUSHED_BACK 3
static int pushed_back[MAX_PUSHED_BACK];
static int num_pushed_back = 0;
static int get_char ()
    {
    if (num_pushed_back)
	{
	int first = pushed_back[0];
	int i;
	for (i = 1; i < num_pushed_back; i++)
	    pushed_back[i-1] = pushed_back[i];
	-- num_pushed_back;
	return first;
	}
    return getchar();
    }
static void push_back (c)
    int                c;
    {
    assert (num_pushed_back < MAX_PUSHED_BACK);
    pushed_back[num_pushed_back++] = c;
    }
static void test_pushback ()
    /* Should act as cat(1)... but slower :-) */
    {
    int			ch;
    int                 npb = 0;
    while ((ch = get_char ()) != EOF) 
	{
	if (npb < MAX_PUSHED_BACK && rand() < RAND_MAX / 2)
	    {
	    push_back (ch);
	    ++ npb;
	    }
	else
	    {
	    (void) putchar (ch);
	    }
	}
    }


static void warn_not_8bit (c)
    char                c;
    {
    if ((unsigned char) c < 0200)
	fprintf(stderr, "warning: converted an escape sequence to "
		"a 7-bit character: %c\n", c);
    }
static void to8bit ()		/* Convert to 8-bit sequences */
    {
    int			ch;	/* Next character read from input */
    int			backch;	/* Backslashed character being built */

    while ((ch = get_char ()) != EOF)
	{
	ch &= 0xFF;
	if (ch != '\\')
	    {
	    /* Not a backslashed sequence at all. */
	    if (ch >= 0200)
		fprintf(stderr, "warning: passing through 8-bit "
			"character unchanged: %c\n", ch);
	    (void) putchar (ch);
	    }
	else
	    {
	    /*
	     * Collect a backslashed character.  If we have to abandon
	     * our reading, because we got a bad character or EOF,
	     * then we output the backslash (since it doesn't start a
	     * legal escape sequence) and push back the remaining
	     * characters for use next time.  This is so that
	     * \\x60 will become \` , for example.
	     */
	    int ch_a, ch_b, ch_c; /* characters after slash */
	    ch_a = get_char ();
	    switch (ch_a)
		{
		case 'x':
		case 'X':
		    /* \x.. hex sequence.  Check first character... */
		    ch_b = get_char ();
		    if (ch_b >= '0'  &&  ch_b <= '9')
			backch = ch_b - '0';
		    else if (ch_b >= 'a'  &&  ch_b <= 'f')
			backch = ch_b - 'a' + 0xA;
		    else if (ch_b >= 'A'  &&  ch_b <= 'F')
			backch = ch_b - 'A' + 0xA;
		    else
			{
			(void) putchar ('\\');
			(void) push_back (ch_a);
			if (ch_b != EOF)
			    (void) push_back (ch_b);
			break;
			}

		    /* Second character. */
		    ch_c = get_char ();
		    if (ch_c >= '0'  &&  ch_c <= '9')
			backch = (backch << 4) | (ch_c - '0');
		    else if (ch_c >= 'a'  &&  ch_c <= 'f')
			backch = (backch << 4) | (ch_c - 'a' + 0xA);
 		    else if (ch_c >= 'A'  &&  ch_c <= 'F')
			backch = (backch << 4) | (ch_c - 'A' + 0xA);
		    else
			{
			(void) putchar ('\\');
			(void) push_back (ch_a);
			(void) push_back (ch_b);
			if (ch_c != EOF)
			    (void) push_back (ch_c);
			break;
			}

		    /* All OK. */
		    warn_not_8bit (backch);
		    (void) putchar (backch);
		    break;
		case '0':
		case '1':
		case '2':
		case '3':
		    /* Octal sequence.  We just got the first
		       character, do the second. */
		    backch = ch_a - '0';
		    ch_b = get_char ();
		    if (ch_b >= '0'  &&  ch_b <= '7')
			backch = (backch << 3) | (ch_b - '0');
		    else
			{
			(void) putchar ('\\');
			(void) push_back (ch_a);
			if (ch_b != EOF)
			    (void) push_back (ch_b);
			break;
			}

		    /* Third character. */
		    ch_c = get_char ();
		    if (ch_c >= '0'  &&  ch_c <= '7')
			backch = (backch << 3) | (ch_c - '0');
		    else
			{
			(void) putchar ('\\');
			(void) push_back (ch_a);
			(void) push_back (ch_b);
			if (ch_c != EOF)
			    (void) push_back (ch_c);
			break;
			}

		    /* All OK. */
		    warn_not_8bit (backch);
		    (void) putchar (backch);
		    break;
		default:
		    /* Something funny, it's not a valid escape
		       sequence anyhow. */
		    (void) putchar ('\\');
		    if (ch_a != EOF)
			(void) push_back (ch_a);
		}
	    }
	}
    }
